
// misc. math functions

#ifndef __JX_MATH
#define __JX_MATH


#include <math.h>
#include <stdio.h>
#include <assert.h>
#include "jxdef.h"

__forceinline double Clamp(double const x, double const minx, double const maxx)
{
	if (x <= minx)
		return minx;
	else if (x >= maxx)
		return maxx;
	else
		return x;
}

__forceinline float Clamp(float const x, float const minx, float const maxx)
{
	if (x <= minx)
		return minx;
	else if (x >= maxx)
		return maxx;
	else
		return x;
}

__forceinline int Clamp(int const x, int const minx, int const maxx)
{
	if (x <= minx)
		return minx;
	else if (x >= maxx)
		return maxx;
	else
		return x;
}

__forceinline double Lerp(double const x, double const y, double const a)
{
	return x + a * (y - x);
}

__forceinline double ExpIntp(double const x, double const y, double const a)
{
	assert((x > 0 && y > 0) || (x < 0 && y < 0));
	double const lx = log(x);
	double const ly = log(y);
	return exp(lx + a * (ly - lx));
}

// maps v from linear oldmin..oldmax -> linear newmin..newmax
__forceinline double Rescale(double const v, double const oldmin, double const oldmax, double const newmin, double const newmax)
{
	assert(oldmin != oldmax);
	return newmin + (v - oldmin) * (newmax - newmin) / (oldmax - oldmin);
}

// maps v from linear oldmin..oldmax -> exponential newmin..newmax
__forceinline double RescaleLin2Exp(double const v, double const oldmin, double const oldmax, double const newmin, double const newmax)
{
	assert(oldmin != oldmax);
	return ExpIntp(newmin, newmax, (v - oldmin) / (oldmax - oldmin));
}


__forceinline double Square(double const x)
{
	return x * x;
}

__forceinline double Cube(double const x)
{
	return x * x * x;
}

__forceinline double log2(double const x)
{
	return log(x) * (1.0 / log(2.0));
}

__forceinline dword log2(dword const x)
{
	return (int)(log2((double)x) + 0.5);    // rounding needed here because of float roundoff errors
}

// faster than pow(2, x)
__forceinline double pow2(double const x)
{
	return exp(x * log(2.0));
}

// -1 ^ x
__forceinline int powminus1(int const x)
{
	return 1 - ((x&1) << 1);
}


__forceinline unsigned int LowestBit(unsigned int const x)
{
	return x & (0-x);		// 0-x instead of -x to remove warning
} 

__forceinline bool IsPow2(unsigned int const x)
{
	return x != 0 && x == LowestBit(x);
}

__forceinline unsigned int CeilPow2Minus1(unsigned int x)
{
	for (unsigned int i=1; i; i <<= 1)
		x |= x >> i;
	
	return x;
}

__forceinline unsigned int CeilPow2(unsigned int const x)
{
	unsigned int const result = CeilPow2Minus1(x) + 1;
	assert(x <= result);
	return result;
}


double const JXPI = 4 * atan(1.0);

__forceinline double cot(double const x)
{
	return 1.0 / tan(x);
}

__forceinline double AmplitudeToDB(double const a)
{
	return log10(a) * 20;
}

__forceinline double DBToAmplitude(double const a)
{
	return pow(10.0, (a * 0.05));
}

__forceinline double ExpStep(double const from, double const to, double const time)
{
	assert((from > 0 && to > 0) || (from < 0 && to < 0));
	assert(time > 0);
	return pow(to / from, 1.0 / time);

	// faster than return exp((log(to) - log(from)) / time);
}
 
__forceinline double ExpTime(double const from, double const to, double const step)
{ 
	assert((from > 0 && to > 0) || (from < 0 && to < 0));
	assert(step > 0);
	return (log(to) - log(from)) / log(step);
}

// x - ptr to first sample
// t - fraction [0..1[
// L - number of samples left to t
// M - number of samples right to t
inline double SamplePeriodic(double const *x, double const t, int const L, int const M)
{
	if (t == 0)
		return x[L];
	
	int const N = L + M;
	
	double const pm1np1 = powminus1(N+1);
	double const oo2n = 1.0 / (2 * N);
	double const pio2n = JXPI * oo2n;

	double sum = 0;
	
	for (int n = -L; n < M; n++)
	{
		double y0 = x[n + L] * powminus1(n);
		double y1 = tan((t - n) * pio2n);
		double y2 = cot((t - n) * pio2n);
		
		sum += y0 * (pm1np1 * y1 + y2);	
	} 
	
	return (sin(JXPI * t) * oo2n) * sum;
}



/*
   This computes an in-place complex-to-complex FFT 
   x and y are the real and imaginary arrays of 2^m points.
   dir =  1 gives forward transform
   dir = -1 gives reverse transform 
*/
inline void FFT(short const dir, long const m, double *x, double *y)
{
	assert(dir == 1 || dir == -1);
	assert(x != NULL);
	assert(y != NULL);
	assert(m >= 0 && m < 30);

	long n,i,i1,j,k,i2,l,l1,l2;
	double c1,c2,tx,ty,t1,t2,u1,u2,z;

	/* Calculate the number of points */
	n = 1 << m;

	/* Do the bit reversal */
	i2 = n >> 1;
	j = 0;
	for (i = 0; i < n-1; i++) 
	{
		if (i < j) 
		{
			tx = x[i];
			ty = y[i];
			x[i] = x[j];
			y[i] = y[j];
			x[j] = tx;
			y[j] = ty;
		}
	  
		k = i2;
		while (k <= j) 
		{  
			j -= k;
			k >>= 1;
		}
		
		j += k;
   }

	/* Compute the FFT */
	c1 = -1.0; 
	c2 = 0.0;
	l2 = 1;

	for (l=0;l<m;l++) 
	{
		l1 = l2;
		l2 <<= 1;
		u1 = 1.0; 
		u2 = 0.0;
	  
		for (j=0;j<l1;j++) 
		{
			for (i=j;i<n;i+=l2) 
			{
				i1 = i + l1;
				t1 = u1 * x[i1] - u2 * y[i1];
				t2 = u1 * y[i1] + u2 * x[i1];
				x[i1] = x[i] - t1; 
				y[i1] = y[i] - t2;
				x[i] += t1;
				y[i] += t2;
			}
		 
			z =  u1 * c1 - u2 * c2;
			u2 = u1 * c2 + u2 * c1;
			u1 = z;
		}

		c2 = sqrt((1.0 - c1) * 0.5);
		
		if (dir == 1) 
			c2 = -c2;

		c1 = sqrt((1.0 + c1) * 0.5);
	}

	/* Scaling for forward transform */
	if (dir == 1) 
	{
		double const oon = 1.0 / n;

		for (i=0; i < n; i++) 
		{
			x[i] *= oon;
			y[i] *= oon;
		}
   }
   
}

__forceinline int FloatToIntChop (const float& f) 
{ 
    int a         = *reinterpret_cast<const int*>(&f);          // take bit pattern of float into a register
    int sign      = (a>>31);                                      // sign = 0xFFFFFFFF if original value is negative, 0 if positive
    int mantissa  = (a&((1<<23)-1))|(1<<23);                      // extract mantissa and add the hidden bit
    int exponent  = ((a&0x7fffffff)>>23)-127;                     // extract the exponent
    int r         = ((dword)(mantissa)<<8)>>(31-exponent);       // ((1<<exponent)*mantissa)>>24 -- (we know that mantissa > (1<<24))
    return ((r ^ (sign)) - sign ) &~ (exponent>>31);                // add original sign. If exponent was negative, make return value 0.
}

__forceinline int FloatToIntFloor( const float& f )
{
	int a   = *reinterpret_cast<const int*>(&f);         // take bit pattern of float into a register
	int sign  = (a>>31);                // sign = 0xFFFFFFFF if original value is negative, 0 if positive
	
	a&=0x7fffffff;                   // we don't need the sign any more
	
	int exponent = (a>>23)-127;               // extract the exponent
	int expsign   = ~(exponent>>31);              // 0xFFFFFFFF if exponent is positive, 0 otherwise
	int imask  = ( (1<<(31-(exponent))))-1;           // mask for true integer values
	int mantissa = (a&((1<<23)-1));              // extract mantissa (without the hidden bit)
	int r   = ((dword)(mantissa|(1<<23))<<8)>>(31-exponent);      // ((1<<exponent)*(mantissa|hidden bit))>>24 -- (we know that mantissa > (1<<24))
	r = ((r & expsign) ^ (sign)) + ((!((mantissa<<8)&imask)&(expsign^((a-1)>>31)))&sign); // if (fabs(value)<1.0) value = 0; copy sign; if (value < 0 && value==(int)(value)) value++;
	return r;
}

__forceinline int FloatToIntCeil (const float& f)
{
	int a   = *reinterpret_cast<const int*>(&f) ^ 0x80000000;      // take bit pattern of float into a register
	int sign  = (a>>31);                // sign = 0xFFFFFFFF if original value is negative, 0 if positive
	
	a&=0x7fffffff;                   // we don't need the sign any more
	
	int exponent = (a>>23)-127;               // extract the exponent
	int expsign   = ~(exponent>>31);              // 0xFFFFFFFF if exponent is positive, 0 otherwise
	int imask  = ( (1<<(31-(exponent))))-1;           // mask for true integer values
	int mantissa = (a&((1<<23)-1));              // extract mantissa (without the hidden bit)
	int r   = ((dword)(mantissa|(1<<23))<<8)>>(31-exponent);      // ((1<<exponent)*(mantissa|hidden bit))>>24 -- (we know that mantissa > (1<<24))
	r = ((r & expsign) ^ (sign)) + ((!((mantissa<<8)&imask)&(expsign^((a-1)>>31)))&sign); // if (fabs(value)<1.0) value = 0; copy sign; if (value < 0 && value==(int)(value)) value++;
	
	return -r;
}

template <class T>
__forceinline void Normalize(T *psamples, dword const numsamples, T const unit)
{
	T maxi = 0;

	for (dword i = 0; i < numsamples; i++)
	{
		T const x = fabs(psamples[i]);

		if (x > maxi)
			maxi = x; 
	}

	T const scale = unit / maxi;

	for (i = 0; i < numsamples; i++)
		psamples[i] *= scale;

}

#endif