// test.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <string.h>
#include <xmmintrin.h>
#include <vector>
#include <float.h>
#include "dspinternal.h"
#include "jxmath.h"


static void DFT(bool const dir, dword const m, double *x1, double *y1, double *x2, double *y2)
{
	
	for (dword i = 0; i < m; i++) 
	{
		x2[i] = 0;
		y2[i] = 0;
		
		double const arg = - (dir ? 1 : -1) * 2.0 * M_PI * i / (double)m;

		double cosarg = 1;
		double sinarg = 0;

		double const ejr = cos(arg);
		double const eji = sin(arg);

		for (dword k = 0; k < m; k++) 
		{
			x2[i] += (x1[k] * cosarg - y1[k] * sinarg);
			y2[i] += (x1[k] * sinarg + y1[k] * cosarg);

			double costmp = cosarg;

			cosarg = cosarg * ejr - sinarg * eji;
			sinarg = costmp * eji + sinarg * ejr;		

		}
	}
	
	/* Copy the data back */
	if (dir)
	{
		for (i = 0; i < m; i++) 
		{
			x1[i] = x2[i] / m;
			y1[i] = y2[i] / m;
		}
	} 
	else 
	{
		for (i = 0; i < m; i++) 
		{
			x1[i] = x2[i];
			y1[i] = y2[i];
		}
	}
	
	for (i = 0; i < m; i++)
	{
		if (fabs(x1[i]) < 1e-14)
			x1[i] = 0;
		
		if (fabs(y1[i]) < 1e-14)
			y1[i] = 0;
		
	}
}

static void DFTShortInput(bool const dir, dword const m, short const *x1, double *x2, double *y2)
{
	for (dword i = 0; i < m; i++) 
	{
		x2[i] = 0;
		y2[i] = 0;
		
		double const arg = - (dir ? 1 : -1) * 2.0 * M_PI * i / (double)m;
		
		double cosarg = 1;
		double sinarg = 0;
		
		double const ejr = cos(arg);
		double const eji = sin(arg);
		
		for (dword k = 0; k < m; k++) 
		{
			x2[i] += x1[k] * cosarg;
			y2[i] += x1[k] * sinarg;
			
			double costmp = cosarg;
			
			cosarg = cosarg * ejr - sinarg * eji;
			sinarg = costmp * eji + sinarg * ejr;		
			
		}
	}
	
	if (dir)
	{
		for (i = 0; i < m; i++) 
		{
			x2[i] /= m;
			y2[i] /= m;
		}
	} 
	
	for (i = 0; i < m; i++)
	{
		if (fabs(x2[i]) < 0.5)
			x2[i] = 0;
		
		if (fabs(y2[i]) < 0.5)
			y2[i] = 0;
		
	}
}

static void DFTFloatOutput(bool const dir, dword const m, double *x1, double *y1, float *output)
{
	for (dword i = 0; i < m; i++) 
	{
		double acc = 0;
		
		double const arg = - (dir ? 1 : -1) * 2.0 * M_PI * i / (double)m;
		
		double cosarg = 1;
		double sinarg = 0;
		
		double const ejr = cos(arg);
		double const eji = sin(arg);
		
		for (dword k = 0; k < m; k++) 
		{
			acc += (x1[k] * cosarg - y1[k] * sinarg);
			
			double costmp = cosarg;
			
			cosarg = cosarg * ejr - sinarg * eji;
			sinarg = costmp * eji + sinarg * ejr;		
			
		}

	
		output[i] = (float)acc;	
	}
	
}



class CFDWave
{
public:
	CFDWave()
	{
		CosCoeff = SinCoeff = NULL;
	}

	~CFDWave()
	{
		Reset();
	}

	void Reset()
	{
		delete[] CosCoeff;
		delete[] SinCoeff;
	}

	void Create(short const *samples, dword const numsamples)
	{
		Reset();
		
		Size = numsamples;
		
		CosCoeff = new double [numsamples];
		SinCoeff = new double [numsamples];
		
		if (IsPow2(numsamples))
		{
			for (dword i = 0; i < numsamples; i++)
			{
				CosCoeff[i] = samples[i];
				SinCoeff[i] = 0;
			}

			FFT(1, log2(numsamples), CosCoeff, SinCoeff);

		}
		else
		{
			DFTShortInput(true, numsamples, samples, CosCoeff, SinCoeff);
		}

		
	}
	
	void Create(short const *samples, dword const numsamples, dword const maxharm)
	{
		CFDWave tw;
		tw.Create(samples, numsamples);

		Reset();

		Size = maxharm * 2;

		CosCoeff = new double [Size];
		SinCoeff = new double [Size];
		memset(CosCoeff, 0, Size * sizeof(double));
		memset(SinCoeff, 0, Size * sizeof(double));
		
		CosCoeff[0] = tw.CosCoeff[0];
		SinCoeff[0] = tw.SinCoeff[0];
		
		for (dword i = 1; i <= maxharm; i++)
		{
			CosCoeff[i] = tw.CosCoeff[i];
			CosCoeff[Size-i] = tw.CosCoeff[tw.Size-i];
			SinCoeff[i] = tw.SinCoeff[i];
			SinCoeff[Size-i] = tw.SinCoeff[tw.Size-i];
		}
		
	}

	void Create(CFDWave const *ps, dword const size)
	{
		Reset();
		
		Size = size;
		CosCoeff = new double [Size];
		SinCoeff = new double [Size];
		memset(CosCoeff, 0, Size * sizeof(double));
		memset(SinCoeff, 0, Size * sizeof(double));
		
		CosCoeff[0] = ps->CosCoeff[0];
		SinCoeff[0] = ps->SinCoeff[0];
		
		for (dword i = 1; i <= ps->Size/2; i++)
		{
			CosCoeff[i] = ps->CosCoeff[i];
			CosCoeff[Size-i] = ps->CosCoeff[ps->Size-i];
			SinCoeff[i] = ps->SinCoeff[i];
			SinCoeff[Size-i] = ps->SinCoeff[ps->Size-i];
		}

	}
		

	void SetHarmonic(dword const i, double _cos, double _sin)
	{
		if (i > Size / 2)
			return;

		CosCoeff[i] = _cos;
		SinCoeff[i] = -_sin;

		if (i > 0)
		{
			CosCoeff[Size-i] = _cos;
			SinCoeff[Size-i] = _sin;
		}

	}

	double GetHarmonicAmplitude(dword const i) const
	{
		if (i > Size / 2)
			return 0;

		return 2 * sqrt(Square(CosCoeff[i]) + Square(SinCoeff[i]));
	}

	void Synthesize(float *output, dword const outsize)
	{
		CFDWave tw;
		tw.Create(this, outsize);

		if (IsPow2(outsize))
		{
			FFT(-1, log2(tw.Size), tw.CosCoeff, tw.SinCoeff);
			for (dword i = 0; i < outsize; i++)
				output[i] = (float)tw.CosCoeff[i];
		}
		else
		{
			DFTFloatOutput(false, tw.Size, tw.CosCoeff, tw.SinCoeff, output);
		}
	}

public:
	dword Size;

	double *CosCoeff;
	double *SinCoeff;
	
};





static void DFTReal(dword const m, double *input, double *outcos, double *outsin)
{
	dword const numout = (m >> 1) + 1;
	

	for (dword i = 0; i < numout; i++) 
	{
		double const arg = -2.0 * JXPI * i / (double)m;
		
		double cosarg = 1;
		double sinarg = 0;
		
		double const ejr = cos(arg);
		double const eji = sin(arg);
		
		double cosacc = 0;
		double sinacc = 0;

		for (dword k = 0; k < m; k++) 
		{
			cosacc += input[k] * cosarg;
			sinacc += input[k] * sinarg;
			
			double costmp = cosarg;
			
			cosarg = cosarg * ejr - sinarg * eji;
			sinarg = costmp * eji + sinarg * ejr;		
			
		}

		outcos[i] = cosacc;
		outsin[i] = sinacc;
		
	}
	
	for (i = 0; i < numout; i++) 
	{
		outcos[i] = outcos[i] / m;
		outsin[i] = outsin[i] / m;
	}

	for (i = 1; i < ((m >> 1) + (m & 1)); i++) 
	{
		outcos[i] = outcos[i] * 2;
		outsin[i] = outsin[i] * -2;
	}
	

	for (i = 0; i < numout; i++)
	{
		if (fabs(outcos[i]) < 1e-14)
			outcos[i] = 0;
		
		if (fabs(outsin[i]) < 1e-14)
			outsin[i] = 0;
		
	}
}




static CXMipWave *CreateMipWave(short const *samples, dword const numsamples, dword const size, float const freq_limit, float const _alias_limit, float const alias_level_limit)
{
	if (!IsPow2(size))
		return NULL;

	if (freq_limit < 0 || freq_limit > 0.5 || _alias_limit < 0 || _alias_limit > 0.5)
		return NULL;

	if (alias_level_limit < 0)
		return NULL;
		

	
	float const alias_limit = 0.5 + (0.5 - _alias_limit);

	CFDWave fdw;
	fdw.Create(samples, numsamples, __min(numsamples / 2, size / 2));

	fdw.SetHarmonic(0, 0, 0);	// remove dc

	int hh = numsamples / 2;

	float maxharmlevel = 0;
	for (dword i = 1; i <= hh; i++)
		maxharmlevel = __max(maxharmlevel, fdw.GetHarmonicAmplitude(i));

	maxharmlevel *= DBToAmplitude(-alias_level_limit);

//	printf("%d\n", hh);

	while(hh > 1)
	{
		if (fdw.GetHarmonicAmplitude(hh) > maxharmlevel)
			break;
			
		hh--;
	}
	
//	printf("%d\n", hh);


	CXMipWaveImp *pmw = new CXMipWaveImp;
	

	double l = 0;

	dword lasthh = hh;

	std::vector<float *> tempwaves;

	do
	{
		double f = hh / (double)size;

		if (l / size >= 0.5)
			break;

		for (dword i = lasthh; i > hh; i--)
			fdw.SetHarmonic(i, 0, 0);
		
		lasthh = hh;
		
		float *output = new float [size];
		fdw.Synthesize(output, size);
		tempwaves.push_back(output);
	
//		printf("%f, %d, ", l / size * 44100, (int)(l * f * 44100));
			
		l = alias_limit / f;

//		printf("%d, %f\n", hh, l);
		
		double const dhh = freq_limit * size / l;
		hh = (int)ceil(dhh);
	
		while (l * hh / (double)size >= 0.5)
			hh--;


		dword const il = l * (1 << (32 - (int)log2(size)));
		
		pmw->Limits.push_back(il);


	} while(1);

	float maks = 0;

	for (i = 0; i < tempwaves.size(); i++)
	{
		for (dword j = 0; j < size; j++)
		{
			if (fabs(tempwaves[i][j]) > maks)
				maks = fabs(tempwaves[i][j]);
		}
	}


	pmw->LevelSize = size;
	pmw->log2LevelSize = log2(size);
	pmw->Data = new short [(pmw->LevelSize+3) * (pmw->Limits.size())];
	short *pout = pmw->Data;
		
	float const ratio = 16383.0f / maks;

	for (i = 0; i < tempwaves.size(); i++)
	{
		for (dword j = 0; j < size; j++)
		{
			*pout++ = ratio * tempwaves[i][j];		
		}

		*pout++ = pout[-size];
		*pout++ = pout[-size];
		*pout++ = pout[-size];

		delete[] tempwaves[i];
	
	}
		
	return pmw;
}



void SetMipWave(CXDSPInterface *p)
{
	p->CreateMipWave = CreateMipWave;
}

dword CXMipWaveImp::QueryMemoryUsage() const
{
	return sizeof(this) + Limits.size() * (LevelSize + 3);
}
