// AudioGraph1.cpp: implementation of the CAudioGraph class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AudioGraphImpl.h"
#include "AudioGraphNodeImpl.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CXAudioGraphImpl::CXAudioGraphImpl()
{
	pCompiler = NULL;
	pInstructions = NULL;
	InputPtrs = OutputPtrs = NULL;
}

CXAudioGraphImpl::~CXAudioGraphImpl()
{
	delete pInstructions;
	delete InputPtrs;	// OutputPtrs is part of this
}

bool CXAudioGraphImpl::AddNode(CXAudioGraphNode *pnode, bool const output_node)
{
	if (pnode->pImpl->GetGraph() == this)
		return false;

	// ownership of the node is transfered to the graph
	Nodes.push_back(std::auto_ptr<CXAudioGraphNode>(pnode));
	pnode->pImpl->SetGraph(this);

	if (output_node)
		OutputNodes.push_back(pnode);

	return true;
}

bool CXAudioGraphImpl::RemoveNode(CXAudioGraphNode *pnode)
{
	if (pnode->pImpl->GetGraph() != this)
		return false;

	for (CAudioGraphNodeList::iterator i = Nodes.begin(); i != Nodes.end(); i++)
	{
		if ((*i).get() == pnode)
		{
			(*i).release();		// ownership is given away
			Nodes.erase(i);
			break;
		}
	}

	for (CAudioGraphNodePtrList::iterator j = OutputNodes.begin(); j != OutputNodes.end(); j++)
	{
		if (*j == pnode)
		{
			OutputNodes.erase(j);
			break;
		}
		
	}
		
	pnode->pImpl->SetGraph(NULL);
	
	return true;
}

bool CXAudioGraphImpl::IsCyclic()
{
	for (CAudioGraphNodeList::iterator i = Nodes.begin(); i != Nodes.end(); i++)
		(*i).get()->pImpl->CycleTestFlag = false;

	for (CAudioGraphNodePtrList::iterator j = OutputNodes.begin(); j != OutputNodes.end(); j++)
	{
		if ((*j)->pImpl->CycleTest())
			return true;
	}

	return false;
}

void CXAudioGraphImpl::SetAudioBufferPool(CXAudioBufferPool *pabp)
{
	pABP = pabp;
}


void CXAudioGraphImpl::Compile()
{
	printf("---- begin compile ----\n");

	assert(pCompiler == NULL);
	pCompiler = new CAGCompiler;
	pCompiler->InstrBuffer = new CAGInstructionBuffer;

	for (CAudioGraphNodeList::iterator i = Nodes.begin(); i != Nodes.end(); i++)
		(*i).get()->pImpl->Processed = false;
	
	for (CAudioGraphNodePtrList::iterator j = OutputNodes.begin(); j != OutputNodes.end(); j++)
		(*j)->pImpl->Compile();

	
	pCompiler->InstrBuffer->AddEnd();

	printf("%d buffers required\n", pCompiler->Buffers.size());
	printf("%d dwords in instruction buffer\n", pCompiler->InstrBuffer->GetSize());
	printf("max in: %d, max out: %d\n", pCompiler->MaxInputs, pCompiler->MaxOutputs);

	{
		CJxCSLock cslock(CS);
		pInstructions = pCompiler->InstrBuffer;
		pCompiler->InstrBuffer = NULL;

		delete[] InputPtrs;
		InputPtrs = new float * [pCompiler->MaxInputs + pCompiler->MaxOutputs];
		OutputPtrs = InputPtrs + pCompiler->MaxInputs;

		Buffers.resize(pCompiler->Buffers.size());
	}

	delete pCompiler;
	pCompiler = NULL;

	printf("---- end compile ----\n");
	
}





void CXAudioGraphImpl::Run(dword const nsamples)
{
	int const *pin = pInstructions->GetPtr();
	
	while(1)
	{
		CAGInstructionBuffer::instruction const inst = (CAGInstructionBuffer::instruction)*pin++;
		if (inst == CAGInstructionBuffer::end)
			break;

		if (inst == CAGInstructionBuffer::alloc_buffer)
		{
			int const ibuf = *pin++;

			assert(ibuf >= 0);
			assert(ibuf < Buffers.size());
			assert(!Buffers[ibuf].Allocated);
			assert(Buffers[ibuf].pBuffer == NULL);
			
			Buffers[ibuf].Allocated = true;
			Buffers[ibuf].pBuffer = pABP->AllocateBuffer();
		}
		else if (inst == CAGInstructionBuffer::free_buffer)
		{
			int const ibuf = *pin++;

			assert(ibuf >= 0);
			assert(ibuf < Buffers.size());
			assert(Buffers[ibuf].Allocated);
			
			Buffers[ibuf].Allocated = false;

			if (Buffers[ibuf].pBuffer != NULL)
			{
				pABP->FreeBuffer(Buffers[ibuf].pBuffer);
				Buffers[ibuf].pBuffer = NULL;
			}

		}
		else if (inst == CAGInstructionBuffer::alloc_copy_buffer)
		{
			int const ibuf = *pin++;
			int const isrc = *pin++;
			
			assert(ibuf >= 0);
			assert(ibuf < Buffers.size());
			assert(!Buffers[ibuf].Allocated);
			assert(Buffers[ibuf].pBuffer == NULL);
			
			Buffers[ibuf].Allocated = true;
			Buffers[ibuf].pBuffer = pABP->AllocateBuffer();

			if (Buffers[isrc].pBuffer != NULL)
				xdsp.Copy_V_V(*Buffers[ibuf].pBuffer, *Buffers[isrc].pBuffer, nsamples);
			else
				xdsp.Clear_V(*Buffers[ibuf].pBuffer, nsamples);

		}
		else if (inst == CAGInstructionBuffer::add_buffer)
		{
			int const ibuf = *pin++;
			int const isrc = *pin++;

			if (Buffers[isrc].pBuffer != NULL)
			{
				if (Buffers[ibuf].pBuffer == NULL)
				{
					Buffers[ibuf].pBuffer = pABP->AllocateBuffer();
					xdsp.Clear_V(*Buffers[ibuf].pBuffer, nsamples);
				}

				xdsp.Add_V_V(*Buffers[ibuf].pBuffer, *Buffers[isrc].pBuffer, nsamples);
			}
			
		}
		else if (inst == CAGInstructionBuffer::callback)
		{
			CXAudioGraphCallback *pcb = (CXAudioGraphCallback *)*pin++;
			AUDIO_GRAPH_CALLBACK_FUNC pfunc = *(AUDIO_GRAPH_CALLBACK_FUNC *)&*pin;
			pin += 2;

			int const numin = *pin++;
			int const *inpind = pin;
			pin += numin;
			int const numout = *pin++;
			int const *outind = pin;
			pin += numout;

			for (int i = 0; i < numin; i++)
			{
				if (inpind[i] == -1)
					InputPtrs[i] = NULL;
				else
				{
					if (Buffers[inpind[i]].pBuffer != NULL)
						InputPtrs[i] = *Buffers[inpind[i]].pBuffer;
					else
						InputPtrs[i] = NULL;
				}
			}

			for (i = 0; i < numout; i++)
			{
				if (outind[i] == -1)
					OutputPtrs[i] = NULL;
				else
				{
					if (Buffers[outind[i]].pBuffer != NULL)
						OutputPtrs[i] = *Buffers[outind[i]].pBuffer;
					else
						OutputPtrs[i] = NULL;
				}
			}
			
			(pcb->*pfunc)(InputPtrs, OutputPtrs, nsamples);

			for (i = 0; i < numout; i++)
			{
				if (outind[i] != -1 && OutputPtrs[i] == NULL)
				{
					pABP->FreeBuffer(Buffers[outind[i]].pBuffer);
					Buffers[outind[i]].pBuffer = NULL;
				}
			}
			
		}
		
	}

}


// ----------------------------------- interface --------------------------------


CXAudioGraph::CXAudioGraph()
{
	pImpl = new CXAudioGraphImpl;
}

CXAudioGraph::~CXAudioGraph()
{
	delete pImpl;
}

void CXAudioGraph::SetAudioBufferPool(CXAudioBufferPool *pabp) { pImpl->SetAudioBufferPool(pabp); }
bool CXAudioGraph::AddNode(CXAudioGraphNode *pnode, bool const output_node) { return pImpl->AddNode(pnode, output_node); }
bool CXAudioGraph::RemoveNode(CXAudioGraphNode *pnode) { return pImpl->RemoveNode(pnode); }
void CXAudioGraph::Compile() { pImpl->Compile(); }
void CXAudioGraph::Run(dword const nsamples) { pImpl->Run(nsamples); }
